#include <cstdio>
#include <string>
#include <cerrno>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>

/* This client does not care about the remote host name,
 * because a local port can not be reserved twice at the same time.
 */

using namespace std;

#include "sizes.h"

static int localport=0;
static int remoteport=0;

static string RespType;
static string AddInfo;

static int Done(int code=0)
{
	printf("%d, %d : %s : %s\r\n", localport, remoteport, RespType.c_str(), AddInfo.c_str());
	if(code)exit(code);
	return code;
}
static int Error(const char *ErrorType = "INVALID-PORT")
{
	int code = -1;
	RespType = "ERROR";
	if(!ErrorType)
	{
		ErrorType = "UNKNOWN-ERROR";
		code = errno;
	}
	AddInfo = ErrorType;
	return Done(code);
}
static DWORD byteswap(DWORD n)
{
	return  ((n&0xFF) << 24)
	      | ((n&0xFF00) << 8)
	      | ((n&0xFF0000) >> 8)
	      | ((n&0xFF000000) >> 24);
}
static int Forward(DWORD host, int lport, int rport)
{
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock < 0)Error(NULL);
	struct sockaddr_in madr;
	memset(&madr, 0, sizeof madr);
	madr.sin_port   = htons(113);
	madr.sin_family = AF_INET;
	madr.sin_addr.s_addr = byteswap(host);
	if(connect(sock, (struct sockaddr *)&madr, sizeof madr) < 0)Error(NULL);
	char Buf[256];
	sprintf(Buf, "%d,%d\n", rport, lport);
	write(sock, Buf, strlen(Buf));
	char c;
	for(; read(sock, &c, 1)==1 && c!=':'; );
	for(RespType=""; read(sock, &c, 1)==1 && c != ':'; )if(RespType.size() || c!=' ')RespType += c;
	for(AddInfo=""; read(sock, &c, 1)==1 && c != '\n'; )if(AddInfo.size() || c!=' ')AddInfo += c;
	return close(sock) ? errno : 0;
}
static int Resolve(void)
{
	FILE *fp;
	char Buf[256];
	if((fp = fopen("/proc/net/ip_masquerade", "rt")) != NULL)
	{
		while(fgets(Buf, 255, fp))
		{
			int l, r, masqport;
			DWORD masqhost;
			char proto[5];
			sscanf(Buf, "%s %X:%X %*X:%X %X", proto, &masqhost, &l, &r, &masqport);
			if(strcmp(proto, "TCP"))continue;
			if(r == remoteport && masqport == localport)
				return Forward(masqhost, r, l);
		}
		fclose(fp);
	}
	if((fp = fopen("/proc/net/tcp", "rt")) != NULL)
	{
		while(fgets(Buf, 255, fp))
		{
			int l, r, uid, state;
			sscanf(Buf, "%*d: %*X:%X %*X:%X %X %*X:%*X %*X:%*X %*X %d", &l, &r, &state, &uid);
		//	if(state != IPPROTO_TCP)continue;
			if(l==localport && r==remoteport)
			{
				const struct passwd *p = getpwuid(uid);
				RespType = "USERID";
				AddInfo = "UNIX : ";
				if(p)
					AddInfo += p->pw_name;
				else
				{
					sprintf(Buf, "%d", uid);
					AddInfo += Buf;
				}
				return 0;
			}
		}
		fclose(fp);
	}
	return Error("NO-USER");
}
int main(int argc, const char *const *argv)
{
	if(argc != 1)
	{
		fprintf(stderr,
		  "Bisqwit's identd version "VERSION"\nCopyright (C) 1992,2000 Bisqwit (http://iki.fi/bisqwit/)\n"
		  "\n"
		  "To be spawned from inetd, without any parameters.\n"
		  "Read the installation instructions from INSTALL.\n\n");
		return -1;
	}
	int l, r, c;

	for(l=0; isdigit(c = getchar()); )l = l*10 + (c-'0');
	while(c == ' ')c = getchar();
	if(c != ',')Error();
	for(r=0; isdigit(c = getchar()) || c==' '; )if(c!=' ')r = r*10 + (c-'0');
	while(c == '\r')c = getchar();
	if(c != '\n' && c != EOF)Error();
	
	remoteport = r;
	localport = l;
	
	return Done(Resolve());
}
